(***********************************************************************)
(*                                                                     *)
(*                               Grimage                               *)
(*                                                                     *)
(*                           Graphics image                            *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2014-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: colors.mli,v 1.9 2014/02/27 16:45:22 fclement Exp $ *)

(** {3 Definitions and manipulations of colors.} *)

type t = Graphics.color;;

(** {6 The RGB color encoding.} *)

(** RGB color encoding, using color decomposition into red, green, and blue
    components. Each component has a value in the range [0..255]. The three
    components are packed together into an integer: [0xRRGGBB], where [RR]
    are the two hexadecimal digits for the red component, [GG] for the green
    component, [BB] for the blue component. *)

type rgb_channel = int;;

type red = rgb_channel
and green = rgb_channel
and blue = rgb_channel
;;

val color_of_rgb : red -> green -> blue -> t;;
 (** [color_of_rgb r g b] returns the integer encoding a color, by packing its
    red component [r], its green component [g], and its blue component [b]. *)
val rgb : red -> green -> blue -> t;;
 (** [rgb] is an alias for [color_of_rgb]. *)

val red_of_color : t -> red;;
val green_of_color : t -> green;;
val blue_of_color : t -> blue;;
val rgb_of_color : t -> red * green * blue;;
 (** Extracting the red, green, and blue components of a color. *)

(** {6 The HSB color encoding.} *)

(** HSB color encoding, using color decomposition into hue, saturation, and
  brightness components.
  Hue is the wavelength within the visible-light spectrum at which the
  energy output from a source of this color is the greatest.
  Saturation measures the purity of colors: high saturation means a
  pure color, low saturation means a faint color.
  Brightness (also called brilliance) is the maximum of the RGB
  components of the color.
  As for RGB encoding, each component has a value in the range [0..255] and the
  three components are packed into an integer. *)

type hsb_channel = int;;

type hue = hsb_channel
and saturation = hsb_channel
and brightness = hsb_channel
;;

val hue_of_color : t -> hue;;
val saturation_of_color : t -> saturation;;
val brightness_of_color : t -> brightness;;
val hsb_of_color : t -> hue * saturation * brightness;;
 (** Decomposition of a color into its hue, saturation, and brightness
  components. *)

val color_of_hsb : hue -> saturation -> brightness -> t;;
 (** [color_of_hsb h s b] returns the color with correspondant HSB
  components. *)
val hsb : hue -> saturation -> brightness -> t;;
 (** [hsb] is an alias for [color_of_hsb]. *)

(** {6 Color encoding conversions.} *)

val hsb_of_rgb : red -> green -> blue -> hue * saturation * brightness;;
val rgb_of_hsb : hue -> saturation -> brightness -> red * green * blue;;
 (** Mappings beween RGB and HSB representations of colors. [hsb_of_rgb] and
   [rgb_of_hsb] are not strictly reciprocal, due to limit conditions, where
   some components are undefined and arbitrarily set to 0. *)

(** {6 Shorthands to access color components.} *)

 (** Abbreviations for accessing color components. *)
val r_of : t -> red;;
val g_of : t -> green;;
val b_of : t -> blue;;
val hue_of : t -> hue;;
val sat_of : t -> saturation;;
val bri_of : t -> brightness;;
 (** Return the red, green, blue, hue, saturation, and brightness of a color. *)
