(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uaref.ml,v 1.33 2014/02/09 17:27:28 fclement Exp $ *)

let ar_qual = Upervasives.qualify "Uaref";;

let ar_failwith name msg =
  Upervasives.failwith_qual ar_qual ""
    (Printf.sprintf "Abstract reference %S: %s" name msg)
;;

(* Abstract flags.*)

type flag =
  {
   get_flag : unit -> bool;
   set_flag : unit -> unit;
   unset_flag : unit -> unit;
  }
;;

let make_flag () =
  let r = ref false in
  {
   get_flag = (fun () -> !r);
   set_flag = (fun () -> r := true);
   unset_flag = (fun () -> r := false);
  }
;;

let make_embedded_flags () =
  let { get_flag = get_specific;
        set_flag = set_specific;
        unset_flag = unset_specific; } = make_flag () in
  let { get_flag = get_general;
        set_flag = set_general;
        unset_flag = unset_general; } = make_flag () in
  let set_both () = set_general (); set_specific () in
  let unset_both () = unset_specific (); unset_general () in
  ({ get_flag = get_specific;
     set_flag = set_specific;
     unset_flag = unset_both; },
   { get_flag = get_general;
     set_flag = set_both;
     unset_flag = unset_general; })
;;

type tflag =
  {
   get_tflag : unit -> bool;
   toggle_tflag : unit -> unit;
  }
;;

let make_toggle_flag b =
  let r = ref b in
  { get_tflag = (fun () -> !r);
    toggle_tflag = (fun () -> r := not !r); }
;;

let make_antonymic_toggle_flags b =
  let { get_tflag = get; toggle_tflag = toggle; } = make_toggle_flag b in
  let set_to b = if not (get ()) = b then toggle () in
  let toggle () = set_to true in
  let get_antonymic () = not (get ()) in
  let toggle_antonymic () = set_to false in
  ({ get_tflag = get; toggle_tflag = toggle; },
   { get_tflag = get_antonymic; toggle_tflag = toggle_antonymic; })
;;

(* Abstract references. *)

type 'a aref =
  {
    get : unit -> 'a;
    set : 'a -> unit;
  }
;;

type 'a t = 'a aref;;

type aref_name = string

and assignment_constraint =
  {
   left : Unatural.t option;
   right : Unatural.t option;
  }

and protection_flag =
  | Not_protected
  | Protected

and 'a bounds_spec =
  {
    min : 'a option;
    max : 'a option;
    compare : 'a -> 'a -> int;
    string_of : 'a -> string;
  }

and 'a bounds =
  | No_bounds
  | Bounds of 'a bounds_spec

and ('a, 'b) aref_spec =
  {
   name : aref_name;
   assignment_constraint : assignment_constraint;
   default : 'a;
   protection_flag : protection_flag;
   bounds : 'b bounds;
  }
;;

type kind =
  | Unconstrained
  | Single
  | Unique
  | Mandatory
;;

let assignment_constraint_of_kind = function
  | Unconstrained -> { left = None; right = None; }
  | Single -> { left = None; right = Some Unatural._1_; }
  | Unique -> { left = Some Unatural._1_; right = Some Unatural._1_; }
  | Mandatory -> { left = Some Unatural._1_; right = None; }
;;

let kind_of_assignment_constraint name = function
  | { left = None; right = None; } -> Unconstrained
  | { left = None; right = Some r; } when r = Unatural._1_ -> Single
  | { left = Some l; right = None; } when l = Unatural._1_ -> Mandatory
  | { left = Some l; right = Some r; }
        when (l, r) = (Unatural._1_, Unatural._1_) -> Unique
  | { left = Some l; right = Some r; } when l = r ->
      ar_failwith name "should be a kind for exactly n assignments with n <> 1"
  | { left = Some _; _ } | { right = Some _; _ } -> assert false
;;

let { get_flag = get_cautious_flag; set_flag = set_cautious_flag;
      unset_flag = unset_cautious_flag; } = make_flag ()
;;

set_cautious_flag ();;

let min_assign_constraint nb_assign ac =
  match ac.left with
  | None -> true
  | Some min -> min <= nb_assign
;;

let max_assign_constraint nb_assign ac =
  match ac.right with
  | None -> true
  | Some max -> nb_assign <= max
;;

let cautious = function
  | Not_protected -> false
  | Protected -> get_cautious_flag ()
;;

let check_bounds = function
  | No_bounds -> (fun _ _ -> ())
  | Bounds { min = min; max = max; compare = cmp; string_of = s_of; } ->
      (fun name i ->
        let check_bound what ord i bound =
          if ord (cmp bound i) 0 then
            ar_failwith name
              (Printf.sprintf
                 "contents must be set to a value %s than or equal to %s, \
                 but value %s is not"
                 what (s_of bound) (s_of i)) in
        let check_bound_opt what ord bound_opt i =
          Upervasives.option (check_bound what ord i) () bound_opt in
        check_bound_opt "greater" (>) min i;
        check_bound_opt "lower" (<) max i)
;;

let check_fetch name nb_assign ac =
  if not (min_assign_constraint nb_assign ac) then
    ar_failwith name "minimum number of assignments not reached"
;;

let check_modif name pf =
  if cautious pf then ar_failwith name "protected from modification"
;;

let check_store name nb_assign ac pf =
  check_modif name pf;
  if not (max_assign_constraint nb_assign ac) then
    ar_failwith name "maximum number of assignments already reached"
;;

let make_aref_gen
    { name = name;
      assignment_constraint = ac;
      default = default;
      protection_flag = pf;
      bounds = bounds; } =
  check_bounds bounds name default;
  let r = ref default in
  let number_of_assignments = ref Unatural._0_ in
  let check_get () = check_fetch name !number_of_assignments ac in
  let check_set x =
    check_store name !number_of_assignments ac pf;
    check_bounds bounds name x in
  { get = (fun () -> check_get (); !r);
    set = (fun x -> check_set x; r := x; Unatural.incr number_of_assignments); }
;;

let aref_spec default =
  {
   name = "";
   assignment_constraint = assignment_constraint_of_kind Unconstrained;
   default = default;
   protection_flag = Not_protected;
   bounds = No_bounds;
  }
;;

let with_name name ({ name = old_name; _ } as aref_spec) =
  if old_name <> "" then
    ar_failwith old_name (Printf.sprintf "cannot be renamed %S" name) else
  { aref_spec with name = name; }
;;

let string_of_kind = function
  | Unconstrained -> "unconstrained"
  | Single -> "single"
  | Unique -> "unique"
  | Mandatory -> "mandatory"
;;

let check_unconstrained_assignment_constraint name ac =
  match kind_of_assignment_constraint name ac with
  | Unconstrained -> ()
  | Single | Unique | Mandatory as k ->
      ar_failwith name
        (Printf.sprintf
           "%s assignment constraint already specified"
           (string_of_kind k))
;;

let check_natural_range name = function
  | { left = Some l; right = Some r; } ->
      if l > r then
        ar_failwith name
          (Printf.sprintf
             "invalid natural range: left = %d is greater than right = %d"
             (Unatural.to_int l) (Unatural.to_int r))
  | _ -> ()
;;

let with_assignment_constraint_gen ac aref_spec =
  check_unconstrained_assignment_constraint aref_spec.name
    aref_spec.assignment_constraint;
  check_natural_range aref_spec.name ac;
  { aref_spec with assignment_constraint = ac; }
;;

let with_single_assignment aref_spec =
  with_assignment_constraint_gen
    (assignment_constraint_of_kind Single) aref_spec

and with_unique_assignment aref_spec =
  with_assignment_constraint_gen
    (assignment_constraint_of_kind Unique) aref_spec

and with_mandatory_assignment aref_spec =
  with_assignment_constraint_gen
    (assignment_constraint_of_kind Mandatory) aref_spec
;;

let with_assignment_constraint l r aref_spec =
  with_assignment_constraint_gen { left = l; right = r; } aref_spec
;;

let with_n_assignments n aref_spec =
  with_assignment_constraint (Some n) (Some n) aref_spec
;;

let single_aref_spec name default =
  with_single_assignment (with_name name (aref_spec default))

and unique_aref_spec name init =
  with_unique_assignment (with_name name (aref_spec init))

and mandatory_aref_spec name init =
  with_mandatory_assignment (with_name name (aref_spec init))
;;

let make_aref default =
  make_aref_gen (aref_spec default)

and make_aref_single name default =
  make_aref_gen (single_aref_spec name default)

and make_aref_unique name init =
  make_aref_gen (unique_aref_spec name init)

and make_aref_mandatory name init =
  make_aref_gen (mandatory_aref_spec name init)
;;

let with_protection
    ({ name = name;
       assignment_constraint = ac;
       protection_flag = pf; _ } as aref_spec) =
  begin
    match kind_of_assignment_constraint name ac with
    | Unconstrained | Single | Mandatory -> ()
    | Unique ->
        ar_failwith name "cannot be protected (unique assignment constraint)"
  end;
  match pf with
  | Protected -> ar_failwith name "protection already specified"
  | Not_protected -> { aref_spec with protection_flag = Protected; }
;;

let check_bounds_spec name
    { min = min; max = max; compare = compare; string_of = s_of; } =
  match (min, max) with
  | (Some min, Some max) ->
      if compare min max > 0 then
        ar_failwith name
          (Printf.sprintf
             "invalid bounds: min = %s is greater than max = %s"
             (s_of min) (s_of max))
  | (_, _) -> ()
;;

let with_bounds bounds_spec
    ({ name = name; bounds = bounds; _ } as aref_spec) =
  check_bounds_spec name bounds_spec;
  match bounds with
  | Bounds _ -> ar_failwith name "bounds already specified"
  | No_bounds -> { aref_spec with bounds = Bounds bounds_spec; }
;;

let int_bounds_spec min max =
  {
   min = min;
   max = max;
   compare = Pervasives.compare;
   string_of = Pervasives.string_of_int;
  }
;;

let with_int_bounds min max = with_bounds (int_bounds_spec min max);;

(* Abstract references to lists. *)

type 'a aref_list =
  {
    get_list : unit -> 'a list;
    clear_list : unit -> unit;
    add_item : 'a -> unit;
  }
;;

let make_aref_list_gen
    { name = name;
      assignment_constraint = ac;
      default = default;
      protection_flag = pf;
      bounds = bounds; } =
  let r = ref [] in
  let check_get () = check_fetch name (Ulist.length !r) ac in
  let check_clear () = check_modif name pf in
  let check_add x =
    check_store name (Ulist.length !r) ac pf;
    check_bounds bounds name x in
  let get_list () = check_get (); Ulist.rev !r in
  let clear_list () = check_clear (); r := [] in
  let add_item x = check_add x; r := x :: !r in
  Ulist.iter add_item (Ulist.rev default);
  { get_list = get_list; clear_list = clear_list; add_item = add_item; }
;;

let make_aref_list default = make_aref_list_gen (aref_spec default);;

let make_aref_list_min_max name default min max =
  make_aref_list_gen
    (with_assignment_constraint (Some min) (Some max)
       (with_name name (aref_spec default)))
;;

let make_aref_list_n name default n = make_aref_list_min_max name default n n;;
